"""
"""

__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

__all__ = ['Application']

import sys
import os
import gettext

import ui
import gtkui
import scene
import game
import chess.board
import ai

import chess.pgn

class Config:
    """
    """
    
    __directory = None
    
    def __init__(self):
        """Constructor for a confgiuration object"""
        self.__directory = os.path.expanduser('~/.glchess')
        
        # Create the directory if it does not exist
        if not os.path.exists(self.__directory):
            os.mkdir(self.__directory)
        else:
            assert(os.path.isdir(self.__directory))

    def getAutosavePath(self):
        """Get the path to the autosave file"""
        return self.__directory + '/autosave.pgn'

class MovePlayer(game.ChessPlayer):
    """This class provides a pseudo-player to watch for piece movements"""
    # The game to control
    __game = None
    
    # A dictionary of pieces added into the scene
    __pieces = None

    def __init__(self, chessGame):
        """Constructor for a move player.
        
        'chessGame' is the game to make changes to (GLChessGame).
        """
        self.__game = chessGame
        game.ChessPlayer.__init__(self, '@move')
        
    # Extended methods

    def onPieceMoved(self, piece, start, end):
        """Called by chess.board.ChessPlayer"""
        self.__game.scene._movePiece(piece, start, end)
        
    def onPlayerMoved(self, player, move):
        """Called by chess.board.ChessPlayer"""
        self.__game._onPlayerMoved(player, move)

class HumanPlayer(game.ChessPlayer):
    """
    """    
    __game = None
    
    def __init__(self, chessGame, name):
        game.ChessPlayer.__init__(self, name)
        self.__game = chessGame
        
    def readyToMove(self):
        self.__game.scene.setHumanPlayer(self)

class Scene(scene.opengl.SceneGL, scene.SceneHumanInput):
    """
    """
    
    __game = None
    
    # TODO
    __moveNumber   = -1
    __pieceModels  = None
    
    # The current human player or None if not a player in play
    __humanPlayer = None

    def __init__(self, chessGame):
        """Constructor for a glChess scene.
        
        'chessGame' is the game the scene is rendering (game.ChessGame).
        """
        self.__game = chessGame
        self.__pieceModels = {}

        # Call parent constructors
        scene.SceneHumanInput.__init__(self)
        scene.opengl.SceneGL.__init__(self)

    def setHumanPlayer(self, player):
        """TODO
        """
        self.__humanPlayer = player
        
        # Animate the board
        if player is self.__game.getWhite():
            self.setBoardRotation(0.0)
        elif player is self.__game.getBlack():
            self.setBoardRotation(180.0)
        else:
            assert(False), 'Human player is not white or black'
            
    def setMoveNumber(self, moveNumber):
        """Set the move number to watch.

        'moveNumber' is the move to watch (integer).
        """
        if self.__moveNumber == moveNumber:
            return
        self.__moveNumber = moveNumber
        
        # Lock the scene if not tracking the game
        self.enableHumanInput(moveNumber == -1)
        
        # Get the state of this scene
        piecesByLocation = self.__game.getAlivePieces(moveNumber)
        
        # Remove any models not present
        requiredPieces = piecesByLocation.values()
        for (piece, model) in self.__pieceModels.items():
            try:
                requiredPieces.index(piece)
            except ValueError:
                self.__pieceModels.pop(piece)
                self.removeChessPiece(model)
        
        # Move the models in the scene
        for (location, piece) in piecesByLocation.iteritems():
            self.__movePiece(piece, location)
            
    def _movePiece(self, piece, start, end):
        """TODO
        """
        # Ignore if not watching the active game
        if self.__moveNumber != -1:
            return
        
        self.__movePiece(piece, end)
        
    def __movePiece(self, piece, location):
        """
        """
        # Get the model for this piece creating one if it doesn't exist
        try:
            model = self.__pieceModels[piece]
        except KeyError:
            # No need to create if didn't exist anyway
            if location is None:
                return
            
            # Make the new model
            pieceName = {chess.board.PAWN: 'pawn', chess.board.ROOK: 'rook', chess.board.KNIGHT: 'knight',
                         chess.board.BISHOP: 'bishop', chess.board.QUEEN: 'queen', chess.board.KING: 'king'}[piece.getType()]
            chessSet = {chess.board.WHITE: 'white', chess.board.BLACK: 'black'}[piece.getColour()]
            model = self.addChessPiece(chessSet, pieceName, location)
            self.__pieceModels[piece] = model
            
        # Delete or move the model
        if location is None:
            self.__pieceModels.pop(piece)
            self.removeChessPiece(model)
        else:
            model.move(location)

    # Extended methods
    
    def onRedraw(self):
        """Called by scene.SceneGL"""
        if self.__game.view is not None:
            self.__game.view.render()

    def playerIsHuman(self):
        """Called by scene.SceneHumanInput"""
        return self.__humanPlayer is not None

    def squareIsFriendly(self, coord):
        """Called by scene.SceneHumanInput"""
        owner = self.__game.getSquareOwner(coord)
        if owner is None:
            return False
        return owner is self.__humanPlayer
    
    def canMove(self, start, end):
        """Called by scene.SceneHumanInput"""
        if self.__humanPlayer is None:
            return False

        return self.__humanPlayer.canMove((start,end)) # FIXME: Promotion type
    
    def moveHuman(self, start, end):
        """Called by scene.SceneHumanInput"""
        player = self.__humanPlayer
        self.__humanPlayer = None
        player.move((start, end, chess.board.QUEEN)) # FIXME: Promotion type
        
class View(gtkui.GtkView):
    """
    """
    
    __application = None
    __game = None
    
    def __init__(self, application, game, name):
        """
        """
        self.__application = application
        self.__game = game
        gtkui.GtkView.__init__(self, application.ui, name)

    def onRender(self):
        """Called by ui.View"""
        self.__game.scene.render()
    
    def onReshape(self, width, height):
        """Called by ui.View"""
        self.__game.scene.reshape(width, height)
    
    def onSelect(self, x, y):
        """Called by ui.View"""
        self.__game.scene.select(x, y)
    
    def onDeselect(self, x, y):
        """Called by ui.View"""
        self.__game.scene.deselect(x, y)
    
    def onSetMoveNumber(self, moveNumber):
        """Called by ui.View"""
        self.__game.scene.setMoveNumber(moveNumber)

    def onClose(self):
        """Called by ui.View"""
        # The user requests the game to end, for now we just do it
        self.__game.remove()

class GLChessGame(game.ChessGame):
    """
    """
    # Link back to the main application
    __application  = None
    
    # The name of the game
    __name         = None
    
    # The scene for this game
    scene          = None
    
    # The view watching this scene
    view            = None
    
    # The players in the game
    __movePlayer   = None
    __aiPlayers    = None
    __humanPlayers = None

    def __init__(self, application, name):
        """Constructor for a chess game.
        
        'application' is a reference to the glChess application.
        'name' is the name of the game (string).
        """
        self.__application = application
        self.__name = name
        self.__aiPlayers = []
        self.__humanPlayers = []
        
        # Call parent constructor
        game.ChessGame.__init__(self)

        # Create a scene to render to
        self.scene = Scene(self)
        self.view = View(application, self, name)
        
        # Watch for piece moves with a player
        self.__movePlayer = MovePlayer(self)
        self.addSpectator(self.__movePlayer)

    def getName(self):
        """Returns the name of the game (string)"""
        return self.__name
    
    def addAIPlayer(self, name, profile):
        """Create an AI player.
        
        'name' is the name of the player to create (string).
        'profile' is the the AI profile to use (ai.AIProfile).
        
        Returns an AI player to use (game.ChessPlayer).
        """
        player = ai.AIPlayer(name, profile)
        self.__aiPlayers.append(player)
        return player
    
    def addHumanPlayer(self, name):
        """Create a human player.
        
        'name' is the name of the player to create.
        
        Returns a human player to use (game.ChessPlayer).
        """
        player = HumanPlayer(self, name)
        self.__humanPlayers.append(player)
        return player

    def playerIsHuman(self, player):
        """Test if a player is human.
        
        'player' is the player to check (game.ChessPlayer).
        
        Returns True if this is a human player in this game otherwise False.
        """
        try:
            if self.__humanPlayers.index(player) < 0:
                return False
            else:
                return True
        except ValueError:
            return False

    def toPGN(self, pgnGame):
        """Write the properties of this game into a PGN game.
        
        'pgnGame' is the game to write into (pgn.PGNGame). All the tags should be unset.
        """
        white = self.getWhite()
        black = self.getBlack()
        
        pgnGame.setTag(pgnGame.PGN_TAG_EVENT, self.getName())
        pgnGame.setTag(pgnGame.PGN_TAG_WHITE, white.getName())
        pgnGame.setTag(pgnGame.PGN_TAG_BLACK, black.getName())

        if isinstance(white, ai.AIPlayer):
            pgnGame.setTag('WhiteAI', white.getProfile().getName())
        if isinstance(black, ai.AIPlayer):
            pgnGame.setTag('BlackAI', black.getProfile().getName())

        moves = self.getMoves()
        lastMove = None
        for m in moves:
            if lastMove is not None:
                pgnGame.addMove(lastMove, m.sanMove)
                lastMove = None
            else:
                lastMove = m.sanMove
        if lastMove is not None:
            pgnGame.addMove(lastMove, None)
            
    def tickAIs(self):
        """Tick the state machines for any AI players.
        
        Call this method at a regular interval.
        """
        for p in self.__aiPlayers:
            p.tick()
            
    def animate(self, timeStep):
        """
        """
        self.scene.animate(timeStep)
            
    def remove(self):
        """Remove this game"""
        self.end()
        self.__application._removeGame(self)
        self.view.close()

    # Private methods
    
    def _onPlayerMoved(self, player, move):
        """FIXME: Rename this
        """
        self.view.addMove(move)

class UI(gtkui.GtkUI):
    """
    """
    
    __application = None
    
    def __init__(self, application):
        """
        """
        self.__application = application
        gtkui.GtkUI.__init__(self)
        
    def onAnimate(self, timeStep):
        """Called by UI"""
        self.__application.animate(timeStep)

    def onGameStart(self, gameName, allowSpectators, whiteName, whiteType, blackName, blackType, moves = None):
        """Called by UI"""
        g = self.__application.addGame(gameName, whiteName, whiteType, blackName, blackType)
        print 'Starting game ' + gameName + ' between ' + whiteName + '(' + str(whiteType) + ') and ' + blackName + '(' + str(blackType) + ')'
        if moves:
            g.start(moves)
        else:
            g.start()
        
    def onGameLoad(self, path, returnResult):
        """Called by UI"""
        try:
            p = chess.pgn.PGN(path, 1)
        except chess.pgn.Error, e:
            self.reportError('Unable to open PGN file ' + path, str(e))
            return
        
        # Use the first game
        pgnGame = p[0]
        
        if returnResult is True:
            whiteAI = pgnGame.getTag('WhiteAI')
            blackAI = pgnGame.getTag('BlackAI')
            msg = ''
            if whiteAI and self.__application.getAIProfile(whiteAI) is None:
                msg += "AI '" + whiteAI + "' is not installed, white player is now human"
                whiteAI = None
            if blackAI and self.__application.getAIProfile(blackAI) is None:
                if msg != '':
                    msg += '\n'
                msg += "AI '" + blackAI + "' is not installed, black player is now human"
                blackAI = None

            self.reportGameLoaded(gameName = pgnGame.getTag(pgnGame.PGN_TAG_EVENT),
                                  whiteName = pgnGame.getTag(pgnGame.PGN_TAG_WHITE),
                                  blackName = pgnGame.getTag(pgnGame.PGN_TAG_BLACK),
                                  whiteAI = whiteAI, blackAI = blackAI,
                                  moves = pgnGame.getMoves())
                                  
            if len(msg) > 0:
                self.reportError('Game modified', msg)
        else:
            g = self.__application.addPGNGame(pgnGame)
            moves = pgnGame.getMoves()
            if moves:
                g.start(moves)
            else:
                g.start()

    def onGameJoin(self, localName, localType, serverHost, serverId, game):
        """Called by UI"""
        print 'Joining game ' + game + ' on server ' + serverHost + ':' + str(serverId) + ' as ' + localName + '(' + str(localType) + ')'
        
    def onGameSave(self, view, path):
        """Called by UI"""
        g = self.__application.getGame(view)
        try:
            f = file(path, 'w')
        except IOError, e:
            self.reportError('Unable to save PGN file ' + path, str(e))
            return
        
        pgnGame = chess.pgn.PGNGame()
        g.toPGN(pgnGame)
            
        lines = pgnGame.getLines()
        for line in lines:
            f.write(line + '\n')
        f.write('\n')
        f.close()
        
    def onQuit(self):
        """Called by UI"""
        self.__application.quit()

class Application:
    """
    """
    
    # The configuration
    __config = None

    # The glChess UI
    ui = None
    
    # The AI types
    __aiProfiles = None
    
    # The games present
    __games = None
    
    def __init__(self):
        """Constructor for glChess application"""
        self.__aiProfiles = {}
        self.__games = []
        
        self.__config = Config()

        self.ui = UI(self)

        self.ui.addTimer(10, self.__tickAIs)
        
    def addAIProfile(self, profile):
        """Add a new AI profile into glChess.
        
        'profile' is the profile to add (ai.AIProfile).
        """
        name = profile.getName()
        assert(self.__aiProfiles.has_key(name) is False)
        self.__aiProfiles[name] = profile
        self.ui.addAIEngine(name)

    def getAIProfile(self, name):
        """Get an installed AI profile.
        
        'name' is the name of the profile to get (string).
        
        Return the profile (ai.AIProfile) or None if it does not exist.
        """
        try:
            return self.__aiProfiles[name]
        except KeyError:
            return None

    def addGame(self, name, whiteName, whiteType, blackName, blackType):
        """Add a chess game into glChess.
        
        'name' is the name of the game (string).
        'whiteName' is the name of the white player (string).
        'whiteType' is the AI profile to use for white (string) or None if white is human.
        'blackName' is the name of the black player (string).
        'blackType' is the AI profile to use for black (string) or None if black is human.
        
        Returns the game object. Use game.start() to start the game.
        """
        # Create the game
        g = GLChessGame(self, name)
        self.__games.append(g)

        msg = ''
        if whiteType is None:
            player = g.addHumanPlayer(whiteName)
        else:
            try:
                profile = self.__aiProfiles[whiteType]
                player = g.addAIPlayer(whiteName, profile)
            except KeyError:
                msg += "AI '" + whiteType + "' is not installed, white player is now human"
                player = g.addHumanPlayer(whiteName)
        g.setWhite(player)

        if blackType is None:
            player = g.addHumanPlayer(blackName)
        else:
            try:
                profile = self.__aiProfiles[blackType]
                player = g.addAIPlayer(blackName, profile)
            except KeyError:
                msg += "AI '" + blackType + "' is not installed, black player is now human"
                player = g.addHumanPlayer(blackName)
        g.setBlack(player)
                
        if len(msg) > 0:
            self.ui.reportError('Game modified', msg)

        return g
    
    def addPGNGame(self, pgnGame):
        """Add a PGN game.
        
        'pgnGame' is the game to add (chess.pgn.PGNGame).
        
        Returns the game object. Use game.start() to start the game.
        """
        return self.addGame(pgnGame.getTag(pgnGame.PGN_TAG_EVENT),
                            pgnGame.getTag(pgnGame.PGN_TAG_WHITE),
                            pgnGame.getTag('WhiteAI'),
                            pgnGame.getTag(pgnGame.PGN_TAG_BLACK),
                            pgnGame.getTag('BlackAI'))
                            
    def addMove(self, view, move):
        # TEMP
        self.ui.addMove(view, move)

    def start(self):
        """Run glChess.
        
        This method does not return.
        """
        # Setup some default AI profiles
        ais = [('GNUchess', 'gnuchess'), ('Sjeng', 'sjeng'), ('Amy', 'Amy'), ('Crafty', 'crafty'), ('Faile', 'faile'), ('Invalid', 'invalid')]
        for (name, executable) in ais:
            self.__detectAI(name, executable)
        
        # Load existing games
        self.__autoload()

        # Start UI (does not return)
        self.ui.run()
        
    def animate(self, timeStep):
        """
        """
        for g in self.__games:
            g.animate(timeStep)

    def quit(self):
        """Quit glChess"""
        # Save any open games
        self.__autosave()
        
        # End current games (will delete AIs etc)
        for game in self.__games[:]:
            game.end()

        # Exit the application
        sys.exit()
        
    # Private methods
    
    def _removeGame(self, g):
        """
        """
        self.__games.remove(g)
    
    def __detectAI(self, name, executable):
        """Add an AI profile if the given executable exists.
        
        'name' is a descriptive name for the AI (string).
        'executable' is the executable to the AI engine (string).
        """
        if os.system('which ' + executable) != 0:
            return
        print 'Detected AI: ' + name
        self.addAIProfile(ai.AIProfile(name, executable))
    
    def __tickAIs(self):
        """Tick the AI players in glChess"""
        for g in self.__games:
            g.tickAIs()
        return True
    
    def __autoload(self):
        """Restore games from the autosave file"""
        path = self.__config.getAutosavePath()
        try:
            p = chess.pgn.PGN(path)
            games = p[:]
        except chess.pgn.Error, e:
            print 'Invalid autoload file ' + path + ': ' + str(e)
            games = []
        except IOError, e:
            print 'Unable to autoload from ' + path + ': ' + str(e)
            games = []
            
        # Delete the file once loaded
        try:
            os.unlink(path)
        except OSError:
            pass

        # Restore each game
        for pgnGame in games:
            g = self.addPGNGame(pgnGame)
            g.start(pgnGame.getMoves())
    
    def __autosave(self):
        """Save any open games to the autosave file"""
        if len(self.__games) == 0:
            return
        
        f = file(self.__config.getAutosavePath(), 'a')
        for g in self.__games:
            pgnGame = chess.pgn.PGNGame()
            g.toPGN(pgnGame)
            
            lines = pgnGame.getLines()
            for line in lines:
                f.write(line + '\n')
            f.write('\n')
        f.close()

if __name__ == '__main__':
    gettext.textdomain('glchess')
    app = Application()
    app.start()
